"""Public entrypoints for the FABRIC FABlib API.

This package exposes the high-level manager (:class:`~fabrictestbed_extensions.fablib.fablib.FablibManager`)
and resource abstractions (slices, nodes, interfaces, components, and network services)
used to build and manage experiments on FABRIC.
"""
